# Backtrace-Connector

Connector service allows to download crashes from multiple datasources and transform them to report that Backtrace understands. By using Backtrace-Connector developers can use one source of truth - Backtrace portal, instead of multiple game console crash portals. List of supported integrations:
- Google Stadia
- Playstation 4
- Switch (Nintendo)
- Xbox

# Development
- Download submodules via command: `git submodule update --recursive`
- `yarn install`/`npm install` to install all dependencies and link all modules
- To build application please use `npm run build`/`yarn build` task. 
- To run application please use `npm run start` 
- service also has Visual Studio Code tasks, that allow you to build/connect with debugger and test service.
- To run tests please use `npm run test` 

# Maintenance 
Connector has available three modes:
* worker: connector service doesn't spawn API integration and is mostly focused on downloading data from data sources. Please use this mode to spawn
    more connector instances than 1,
* api: connector service spawns API and worker orchestration mode. In this mode, API is available but to process jobs, connector needs to spawn workers
* hybrid: both - worker and api mdoes together.

*Important notice*
To use connector with background workers, postgres database support is required! You might experience unexpected issues if you run the connector service
with sqlite3 support instead! 

To modify mode, please change the variable "mode" availble in the connector configuration. By default the hybrid mode is enabled. 
To check possible options please check available configuraiton files in the connector solution. 

# Test environment:
* Node 16.14.2
* servers: yolo, connector-test, connector

# Connector overview 

List of connection types with basic information about them

## Xbox

Xbox integration uses Xbox Crash Portal to fetch diagnostic data. For Backtrace Xbox users - please recommend to use SDK instead of using Xbox Crash Portal integration.

### Authentication 

Xbox integration uses Bearer authentication. To receive authentication token, connector has to send user name (user) and user secret (secret) to Microsoft login API. By default, Xbox API will make a call into xbox specific API, but this call requires fetching authentication token from somewhere else.

Authentication token expires after 2 hours. Xbox connection will refresh authentication token after 1.5h. 

### Communication
Xbox integration makes a request to `https://manage.devcenter.microsoft.com` endpoint to fetch specific diagnostic information. Connection doesn't support different server url - user can't override default server url. Authentication token received from Microsoft login API, will grant access only to this specific part of Microsoft API. 

Connector by default, every 3 minutes (`_serviceInterval` variable), will try to fetch new data from Xbox API. If current processing date isn't today, Connector will try to fetch data day by day, until current processing day will be current date. 

To download diagnostic data, connector will create base url `https://manage.devcenter.microsoft.com/v1.0/my/analytics`. To fetch new crashes from Xbox API connector will use `failureHits` endpoint. Data returned from the API will be aggregated by hours - `aggregationLevel` parameter. Also, to extend data provided by default by Microsoft API, we will use `groupby` parameter and extend crash information with:

```
date,failureName,failureHash,eventType,symbol,osVersion,osRelease,deviceType,market,packageVersion,osArchitecture,praid,sandboxId
```

Some of those parameters are not docummented in Microsoft API - for example `packageVersion` or `osArchitecture`. Grouping data by all those parameters, allows us to get the most unique report possible. Otherwise, all reports will be grouped by failureHash. Connector should *always* use startDate and endDate to fetch data from xbox api by date range. Without it, all data will be groupped into a single report. Xbox API also allows to filter diagnostic data - in our scenario, we need to make sure we will download useful data. In this case - we filtered data by using eventType. Xbox API won't return any detailed information about hangs or other events different than crashes. Because of that eventType should be always equal to 'crash'. We wanted also to filter all reports with empty failureHash - unfortunately, API sometimes returned 500 status code, which suggests problems on the server side wtih filter implementation.

`failureHits` endpoint will provide aggregated by hour crash information, however those information don't have an information about cabinet (archive with memory dump), stack trace etc. By using `failureHash` returned by `failureHits` api we're able to download failure details from `failureDetails` endpoint. Unfortunately, groupped data returned from Microsoft API, will assign the same failureHash to multiple the same entries. `failureDetails` endpoint also may return empty response - in case if there is no `failureDetails` in specific time period for crash. With tha being said - Connector will try to grab the last failure detail from `failureDetails` API by using `endDate`, ordering data by `date desc` and grabbing the last one item from API - `top 1`. Because Xbox API updates rarely `failureDetails` we will download only one failure details for specific `failureHash` per day and cache it. Because the same `failureHash` may be assigned to multiple different crashes, we're using details information only to fetch stack trace and cabinet.

`failureDetails` endpoint returns `cabId` - guid that allows Connector to fetch stack trace and cabinet.`stackTrace` endpoint allows to return stack trace for specific failure details. Connector will also use `cabdownload` to download cabinet file with minidump file and other crash information. In case if cabinet already exists on the hard drive - connector cache directory, algorithm will prevent from downloading cabinet file once again.

Microsoft API often fails - server constatly return 500. Every xbox api call return remained-rate-limit header, which will tell us how many requests connector still can do to Microsoft API. Looks like this limit is always set to 1 - which means that we can make 1 request per 5/10sec. Every method responsible for downloading data from Xbox API, will try to retry request at least 3 times.


### Auto recovery

Let's think about scenario - we need to restart connector service. To prevent loosing progress and sending duplicated data to Backtrace after restart, Connector store actual download progres in API cache - XboxCrashCache. Every time when algorithm ends sending data to Backtrace, cache class will update cache JSON file with progress information. On service restart - `XboxCrashCache` will try to read this file - if file exists, then algorithm will add data from JSON to cache. Otherwise cache is empty. 

Connector after processing single data will also update `lastEventId` connection property - by using this property we will know where to start after big failure with connector cache on hard drive. 

